/*
 *	dphttpd - a simple server implementation to test event polling methods
 *
 *			  "this software is a mess"(tm)
 *
 *	this should work to build it :
 *
 *
 *	07-01-2001	by Davide Libenzi <davidel@xmailserver.org>
 *
 */

/* this is stupid to say but only one of these must be defined */


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdarg.h>
#include <string.h>
#include <assert.h>
#include <limits.h>
#include <ctype.h>
#include <time.h>
#include <errno.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sched.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/select.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <arpa/nameser.h>
#include <netdb.h>
#include <syslog.h>
#include <glob.h>
#include <semaphore.h>
#include "coro.h"

/* #include <sys/epoll.h> */
#include "epoll.h"

#include "dbllist.h"




#define CO_STD_STACK_SIZE		(2 * 4096)
#define STD_SCHED_TIMEOUT		1
/* you might need to increase "net.ipv4.tcp_max_syn_backlog" to use this value */
#define STD_LISTEN_SIZE			2048
#define DATA_BUFFER_SIZE		2048
#define MIN_AHEAD_SPACE			(DATA_BUFFER_SIZE / 12)
#define STD_MESSAGE_SIZE		128
#define STD_SERVER_PORT			8080
#define MAX_DEFAULT_FDS			20000



struct dph_conn {
	struct list_head lnk;
	int sfd;
	int events, revents;
	struct coroutine *co;
	char buffer[DATA_BUFFER_SIZE];
	int nbytes, rindex;
};




static int kdpfd;
static struct epoll_event *events;
static int maxfds, numfds = 0;
static int chash_size;
static struct list_head *chash;
static int msgsize = STD_MESSAGE_SIZE, port = STD_SERVER_PORT,
	maxsfd = MAX_DEFAULT_FDS, stksize = CO_STD_STACK_SIZE;
struct sockaddr_in saddr;
static volatile unsigned long httpresp = 0;
static int nreqsess = 1;
static char httpreq[512] = "";



int dph_socket(int domain, int type, int protocol)
{
	int sfd = socket(domain, type, protocol),
		flags = 1;

	if (sfd == -1)
		return -1;
	if (ioctl(sfd, FIONBIO, &flags) &&
	    ((flags = fcntl(sfd, F_GETFL, 0)) < 0 ||
	     fcntl(sfd, F_SETFL, flags | O_NONBLOCK) < 0)) {
		close(sfd);
		return -1;
	}
	return sfd;
}


int dph_close(int sfd)
{
	close(sfd);
	return 0;
}


int dph_connect(struct dph_conn *conn, const struct sockaddr *serv_addr, socklen_t addrlen)
{
	if (connect(conn->sfd, serv_addr, addrlen) == -1) {
		if (errno != EWOULDBLOCK && errno != EINPROGRESS)
			return -1;
		conn->events = EPOLLOUT | EPOLLERR | EPOLLHUP;
		co_resume(conn);
		if (conn->revents & (EPOLLERR | EPOLLHUP))
			return -1;
	}
	return 0;
}


int dph_read(struct dph_conn *conn, char *buf, int nbyte)
{
	int n;

	while ((n = read(conn->sfd, buf, nbyte)) < 0) {
		if (errno == EINTR)
			continue;
		if (errno != EAGAIN && errno != EWOULDBLOCK)
			return -1;
		conn->events = EPOLLIN | EPOLLERR | EPOLLHUP;
		co_resume(conn);
	}
	return n;
}


int dph_write(struct dph_conn *conn, char const *buf, int nbyte)
{
	int n;

	while ((n = write(conn->sfd, buf, nbyte)) < 0) {
		if (errno == EINTR)
			continue;
		if (errno != EAGAIN && errno != EWOULDBLOCK)
			return -1;
		conn->events = EPOLLOUT | EPOLLERR | EPOLLHUP;
		co_resume(conn);
	}
	return n;
}


int dph_accept(struct dph_conn *conn, struct sockaddr *addr, int *addrlen)
{
	int sfd, flags = 1;

	while ((sfd = accept(conn->sfd, addr, (socklen_t *) addrlen)) < 0) {
		if (errno == EINTR)
			continue;
		if (errno != EAGAIN && errno != EWOULDBLOCK)
			return -1;
		conn->events = EPOLLIN | EPOLLERR | EPOLLHUP;
		co_resume(conn);
	}
	if (ioctl(sfd, FIONBIO, &flags) &&
	    ((flags = fcntl(sfd, F_GETFL, 0)) < 0 ||
	     fcntl(sfd, F_SETFL, flags | O_NONBLOCK) < 0)) {
		close(sfd);
		return -1;
	}
	return sfd;
}


static int dph_new_conn(int sfd, void *func)
{
	struct dph_conn *conn = (struct dph_conn *) malloc(sizeof(struct dph_conn));
	struct epoll_event ev;

	if (!conn)
		return -1;

	memset(conn, 0, sizeof(*conn));
	DBL_INIT_LIST_HEAD(&conn->lnk);
	conn->sfd = sfd;
	conn->events = EPOLLIN | EPOLLOUT | EPOLLERR | EPOLLHUP;
	conn->revents = 0;
	conn->nbytes = conn->rindex = 0;
	if (!(conn->co = co_create(func, NULL, stksize))) {
		free(conn);
		return -1;
	}

	DBL_LIST_ADDT(&conn->lnk, &chash[sfd % chash_size]);

	ev.events = EPOLLIN | EPOLLOUT;
	ev.data.ptr = conn;
	if (epoll_ctl(kdpfd, EPOLL_CTL_ADD, sfd, &ev) < 0) {
		fprintf(stderr, "epoll set insertion error: fd=%d\n", sfd);

		DBL_LIST_DEL(&conn->lnk);
		co_delete(conn->co);
		free(conn);
		return -1;

	}

	++numfds;

	co_call(conn->co, conn);

	return 0;
}


static void dph_exit_conn(struct dph_conn *conn)
{
	struct epoll_event ev;

	if (epoll_ctl(kdpfd, EPOLL_CTL_DEL, conn->sfd, &ev) < 0) {
		fprintf(stderr, "epoll set deletion error: fd=%d\n", conn->sfd);

	}

	DBL_LIST_DEL(&conn->lnk);
	dph_close(conn->sfd);
	free(conn);

	--numfds;

	co_exit(0);
}


static int dph_create_conn(int domain, int type, int protocol, void *func)
{
	int sfd = dph_socket(domain, type, protocol);

	return sfd != -1 ? dph_new_conn(sfd, func): -1;
}


static int dph_read_data(struct dph_conn *conn)
{
	int nbytes;

	if (conn->rindex > (sizeof(conn->buffer) - MIN_AHEAD_SPACE)) {
		memcpy(conn->buffer, conn->buffer + conn->rindex, conn->nbytes - conn->rindex);
		conn->nbytes = conn->nbytes - conn->rindex;
		conn->rindex = 0;
	} else if (conn->nbytes == sizeof(conn->buffer) && conn->rindex > 0) {
		if (conn->rindex > (sizeof(conn->buffer) >> 1))
			memcpy(conn->buffer, conn->buffer + conn->rindex, conn->nbytes - conn->rindex);
		else
			memmove(conn->buffer, conn->buffer + conn->rindex, conn->nbytes - conn->rindex);
		conn->nbytes = conn->nbytes - conn->rindex;
		conn->rindex = 0;
	}

	if ((nbytes = dph_read(conn, conn->buffer + conn->rindex,
			       sizeof(conn->buffer) - conn->rindex)) <= 0)
		return -1;

	conn->nbytes += nbytes;

	return 0;
}


static char *dph_read_line(struct dph_conn *conn)
{
	char *nline = NULL;

	for (; !nline;) {
		if (conn->nbytes > conn->rindex) {
			if ((nline = memchr(conn->buffer + conn->rindex, '\n', conn->nbytes - conn->rindex))) {
				char *line = conn->buffer + conn->rindex;

				conn->rindex += (nline - line) + 1;
				for (; nline > line && nline[-1] == '\r'; nline--);
				*nline = '\0';
				return line;
			}
			if (conn->nbytes == sizeof(conn->buffer))
				break;
		}
		if (dph_read_data(conn) < 0)
			break;
	}
	return NULL;
}


static int dph_parse_request(struct dph_conn *conn)
{
	char *line;

	if (!(line = dph_read_line(conn)))
		return -1;


	for (;;) {
		if (!(line = dph_read_line(conn)))
			return -1;
		if (*line == '\0')
			break;

	}
	return 0;
}


static int dph_send_response(struct dph_conn *conn)
{
	static int resplen = -1, wbytes, wcurr;
	static char *resp = NULL;

	if (resp == NULL) {
		msgsize = ((msgsize + 63) / 64) * 64;

		resp = (char *) malloc(msgsize + 256);

		sprintf(resp,
			"HTTP/1.1 200 OK\r\n"
			"Server: dp server\r\n"
			"Content-Type: text/plain\r\n"
			"Content-Length: %d\r\n"
			"\r\n", msgsize);

		while (msgsize > 0) {
			strcat(resp, "01234567890123\r\n" "01234567890123\r\n" "01234567890123\r\n" "01234567890123\r\n");
			msgsize -= 64;
		}

		resplen = strlen(resp);
	}

	for (wbytes = 0; wbytes < resplen;) {
		wcurr = dph_write(conn, resp + wbytes, resplen - wbytes);
		if (wcurr <= 0)
			return -1;
		wbytes += wcurr;
	}

	return 0;
}


static void *dph_httpd(void *data)
{
	struct dph_conn *conn = (struct dph_conn *) data;

	while (dph_parse_request(conn) == 0) {
		dph_send_response(conn);

	}

	dph_exit_conn(conn);
	return data;
}


static void *dph_acceptor(void *data)
{
	struct dph_conn *conn = (struct dph_conn *) data;
	struct sockaddr_in addr;
	int sfd, addrlen = sizeof(addr);

	while ((sfd = dph_accept(conn, (struct sockaddr *) &addr, &addrlen)) != -1) {
		if (dph_new_conn(sfd, dph_httpd) < 0) {
			dph_close(sfd);

		}
	}
	return data;
}


static struct dph_conn *dph_find(int sfd)
{
	struct list_head *head = &chash[sfd % chash_size], *lnk;
	struct dph_conn *conn;

	DBL_LIST_FOR_EACH(lnk, head) {
		conn = DBL_LIST_ENTRY(lnk, struct dph_conn, lnk);

		if (conn->sfd == sfd)
			return conn;
	}
	return NULL;
}


unsigned long long dph_mstics(void)
{

	struct timeval tv;

	if (gettimeofday(&tv, NULL) != 0)
		return (0);

	return (1000 * (unsigned long long) tv.tv_sec + (unsigned long long) tv.tv_usec / 1000);

}



int dph_init(void)
{
	int ii;

	if (!(events = (struct epoll_event *) malloc(maxsfd * sizeof(struct epoll_event)))) {
		perror("malloc()");
		return -1;
	}

	if ((kdpfd = epoll_create(maxsfd)) < 0) {
		perror("epoll_create");
		return -1;
	}

	if (!(chash = (struct list_head *) malloc(maxsfd * sizeof(struct list_head)))) {
		perror("malloc()");
		free(events);
		close(kdpfd);
		return -1;
	}

	maxfds = maxsfd;
	chash_size = maxfds;
	for (ii = 0; ii < maxfds; ii++)
		DBL_INIT_LIST_HEAD(&chash[ii]);

	return 0;
}

int dph_cleanup(void)
{
	free(events);
	close(kdpfd);

	return 0;
}



static int dph_scheduler(int loop, unsigned int timeout)
{
	int ii, nfds;
	struct dph_conn *conn;
	struct epoll_event *cevents;

	do {
		nfds = epoll_wait(kdpfd, events, maxfds, timeout * 1000);

		for (ii = 0, cevents = events; ii < nfds; ii++, cevents++) {
			conn = cevents->data.ptr;
			conn->revents = cevents->events;

			if (conn->revents & conn->events)
				co_call(conn->co, conn);
		}
	} while (loop);
	return 0;
}





#if defined(DPHTTPD)

void dph_usage(char const *prgname)
{

	fprintf(stderr,
		"use: %s [--msgsize nbytes (%d)] [--port nbr (%d)] [--maxfds nfds (%d)]\n\t[--stksize bytes (%d)]\n",
		prgname, msgsize, port, maxsfd, stksize);

}


int main(int argc, char *argv[])
{
	int ii, sfd, flags = 1;
	struct linger ling = {0, 0};
	struct sockaddr_in addr;

	for (ii = 1; ii < argc; ii++) {
		if (strcmp(argv[ii], "--msgsize") == 0) {
			if (++ii < argc)
				msgsize = atoi(argv[ii]);
			continue;
		}
		if (strcmp(argv[ii], "--port") == 0) {
			if (++ii < argc)
				port = atoi(argv[ii]);
			continue;
		}
		if (strcmp(argv[ii], "--maxfds") == 0) {
			if (++ii < argc)
				maxsfd = atoi(argv[ii]);
			continue;
		}
		if (strcmp(argv[ii], "--stksize") == 0) {
			if (++ii < argc)
				stksize = atoi(argv[ii]);
			continue;
		}

		dph_usage(argv[0]);
		return 1;
	}

	if (dph_init() == -1) {

		return 2;
	}

	if ((sfd = dph_socket(AF_INET, SOCK_STREAM, 0)) == -1) {

		dph_cleanup();
		return 3;
	}
	setsockopt(sfd, SOL_SOCKET, SO_REUSEADDR, &flags, sizeof(flags));
	setsockopt(sfd, SOL_SOCKET, SO_KEEPALIVE, &flags, sizeof(flags));
	setsockopt(sfd, SOL_SOCKET, SO_LINGER, &ling, sizeof(ling));

	addr.sin_family = AF_INET;
	addr.sin_port = htons(port);
	addr.sin_addr.s_addr = htonl(INADDR_ANY);
	if (bind(sfd, (struct sockaddr *) &addr, sizeof(addr)) == -1) {

		dph_close(sfd);
		dph_cleanup();
		return 4;
	}

	listen(sfd, STD_LISTEN_SIZE);

	if (dph_new_conn(sfd, (void *) dph_acceptor) == -1) {

		dph_close(sfd);
		dph_cleanup();
		return 5;
	}

	dph_scheduler(1, STD_SCHED_TIMEOUT);

	dph_cleanup();
	return 0;
}

#endif	/* #if defined(DPHTTPD) */





#if defined(HTTP_BLASTER)

static void *dph_http_session(void *data)
{
	int i;
	struct dph_conn *conn = (struct dph_conn *) data;

	if (dph_connect(conn, (struct sockaddr *) &saddr, sizeof(saddr)) == 0) {
		for (i = 0; i < nreqsess; i++) {
			if (dph_write(conn, httpreq, strlen(httpreq)) > 0) {
				static char const *clent = "Content-Length:";
				int length = -1, clens = strlen(clent);
				char *line;
				static char buf[2048];

				while ((line = dph_read_line(conn))) {
					if (*line == '\0')
						break;
					if (strncasecmp(line, clent, clens) == 0)
						length = atoi(line + clens + 1);
				}
				if (length >= 0) {
					if ((conn->nbytes - conn->rindex) > 0)
						length -= conn->nbytes - conn->rindex;
					++httpresp;
				}
				while (length > 0) {
					int rlen = length > sizeof(buf) ? sizeof(buf): length;

					if ((rlen = dph_read(conn, buf, rlen)) <= 0)
						goto sess_out;
					length -= rlen;
				}
			}
			else
				goto sess_out;
		}
	}
sess_out:
	dph_exit_conn(conn);
	return data;
}


void dph_usage(char const *prgname)
{

	fprintf(stderr, "use: %s  --server serv	 --port nprt  --numconns ncon  [--nreq nreq (%d)]\n"
		"[--url url ('/')]	[--stksize bytes (%d)]\n",
		prgname, nreqsess, stksize);

}


int main(int argc, char *argv[])
{
	int ii, nconns = 0, totconns = 0;
	unsigned long resplast;
	unsigned long long tinit, tlast;
	struct hostent * he;
	char const *server = NULL, *url = "/";
	struct in_addr inadr;

	for (ii = 1; ii < argc; ii++) {
		if (strcmp(argv[ii], "--server") == 0) {
			if (++ii < argc)
				server = argv[ii];
			continue;
		}
		if (strcmp(argv[ii], "--port") == 0) {
			if (++ii < argc)
				port = atoi(argv[ii]);
			continue;
		}
		if (strcmp(argv[ii], "--numconns") == 0) {
			if (++ii < argc) {
				nconns = atoi(argv[ii]);
				if (nconns > maxsfd)
					maxsfd = nconns + nconns >> 1 + 1;
			}
			continue;
		}
		if (strcmp(argv[ii], "--nreq") == 0) {
			if (++ii < argc)
				nreqsess = atoi(argv[ii]);
			continue;
		}
		if (strcmp(argv[ii], "--url") == 0) {
			if (++ii < argc)
				url = argv[ii];
			continue;
		}
		if (strcmp(argv[ii], "--stksize") == 0) {
			if (++ii < argc)
				stksize = atoi(argv[ii]);
			continue;
		}

		dph_usage(argv[0]);
		return 1;
	}

	if (!server || !nconns) {
		dph_usage(argv[0]);
		return 2;
	}

	sprintf(httpreq,
		"GET %s HTTP/1.1\r\n"
		"Host: %s\r\n"
		"Connection: keepalive\r\n"
		"\r\n",
		url, server);

	if (inet_aton(server, &inadr) == 0)
	{
		if ((he = gethostbyname(server)) == NULL)
		{
			fprintf(stderr, "unable to resolve: %s\n", server);
			return (-1);
		}

		memcpy(&inadr.s_addr, he->h_addr_list[0], he->h_length);
	}
	saddr.sin_family = AF_INET;
	saddr.sin_port = htons(port);
	memcpy(&saddr.sin_addr, &inadr.s_addr, 4);

	if (dph_init() == -1) {

		return 2;
	}

	resplast = 0;
	tinit = tlast = dph_mstics();

	for (;;) {
		int nfds = numfds, errs = 0, diffconns = nconns - numfds;
		unsigned long long tcurr;

		while (numfds < nconns) {
			dph_create_conn(AF_INET, SOCK_STREAM, 0, dph_http_session);
			if (nfds == numfds) {
				++errs;
				if (errs > 32) {
					fprintf(stderr, "unable to connect: server=%s errors=%d\n", server, errs);
					goto main_exit;
				}
			}
			else
				++totconns;
			nfds = numfds;
		}

		dph_scheduler(0, STD_SCHED_TIMEOUT);

		tcurr = dph_mstics();
		if ((tcurr - tlast) > 1000) {
			printf("rate = %lu	avg = %lu  totconns = %d  diff = %d\n",
			       (unsigned long) ((1000 * (httpresp - resplast)) / (tcurr - tlast)),
			       (unsigned long) ((1000 * httpresp) / (tcurr - tinit)), totconns, diffconns);

			tlast = tcurr;
			resplast = httpresp;
		}
	}

main_exit:
	dph_cleanup();
	return 0;
}

#endif	/* #if defined(HTTP_BLASTER) */





#if defined(PIPETESTER)


int dph_createcgi(char **args, void *func)
{
	int fds[2], flags = 1;
	pid_t chpid;

	if (pipe(fds)) {
		perror("pipe");
		return -1;
	}
	chpid = fork();
	if (chpid == -1) {
		perror("fork");
		close(fds[0]), close(fds[1]);
		return -1;
	} else if (chpid == 0) {
		close(fds[0]);
		dup2(fds[1], 1);
		close(fds[1]);
		execvp(args[0], args);
		perror("exec");
		exit(1);
	}
	close(fds[1]);
	if (ioctl(fds[0], FIONBIO, &flags) &&
	    ((flags = fcntl(fds[0], F_GETFL, 0)) < 0 ||
	     fcntl(fds[0], F_SETFL, flags | O_NONBLOCK) < 0)) {
		close(fds[0]);
		return -1;
	}
	fprintf(stdout, "child-run=%d  fd=%d\n", chpid, fds[0]), fflush(stdout);
	return dph_new_conn(fds[0], func);
}


int dph_createpipetest(int size, int tsleep, int ttime, void *func)
{
	int fds[2], flags = 1;
	pid_t chpid;

	if (pipe(fds)) {
		perror("pipe");
		return -1;
	}
	chpid = fork();
	if (chpid == -1) {
		perror("fork");
		close(fds[0]), close(fds[1]);
		return -1;
	} else if (chpid == 0) {
		int ii;
		char *buff = (char *) malloc(size + 1);
		close(fds[0]);
		dup2(fds[1], 1);
		close(fds[1]);

		srand(getpid() * time(NULL));
		for (ii = 0; ii < (size - 1); ii++) {
			if (ii && !(ii % 64))
				buff[ii] = '\n';
			else
				buff[ii] = '0' + (rand() % 10);
		}
		buff[ii++] = '\n';
		buff[ii] = '\0';
		ttime += (ttime * rand()) / RAND_MAX - (ttime >> 1);
		ttime *= 1000;
		while (ttime > 0) {
			usleep(tsleep * 1000);
			fputs(buff, stdout), fflush(stdout);
			ttime -= tsleep;
		}
		free(buff);
		exit(0);
	}
	close(fds[1]);
	if (ioctl(fds[0], FIONBIO, &flags) &&
	    ((flags = fcntl(fds[0], F_GETFL, 0)) < 0 ||
	     fcntl(fds[0], F_SETFL, flags | O_NONBLOCK) < 0)) {
		close(fds[0]);
		return -1;
	}
	fprintf(stdout, "child-run=%d  fd=%d\n", chpid, fds[0]), fflush(stdout);
	return dph_new_conn(fds[0], func);
}


static void *dph_pipe_session(void *data)
{
	struct dph_conn *conn = (struct dph_conn *) data;
	int nbytes, totbytes = 0;
	char buff[257];

	while ((nbytes = dph_read(conn, buff, sizeof(buff))) > 0) {
		fprintf(stdout, "[%p] %d bytes readed\n", conn, nbytes), fflush(stdout);
		totbytes += nbytes;
	}
	fprintf(stdout, "[%p] exit - totbytes=%d\n", conn, totbytes), fflush(stdout);
	dph_exit_conn(conn);
	return data;
}


void dph_sigchld(int sig)
{
	int status;
	pid_t pid;

	while ((pid = waitpid(0, &status, WNOHANG)) > 0) {
		fprintf(stdout, "child-dead=%d\n", pid), fflush(stdout);
	}
	signal(SIGCHLD, dph_sigchld);
}


void dph_usage(char const *prgname)
{

	fprintf(stderr, "use: %s  [--ncgis ncgi]  [--cgi cgi] [--stksize bytes (%d)]\n",
		prgname, stksize);

}


int main(int argc, char *argv[])
{
	int ii, ncgis = 8;
	char *cgi = NULL;
	char *args[16];

	for (ii = 1; ii < argc; ii++) {
		if (strcmp(argv[ii], "--ncgis") == 0) {
			if (++ii < argc)
				ncgis = atoi(argv[ii]);
			continue;
		}
		if (strcmp(argv[ii], "--cgi") == 0) {
			if (++ii < argc)
				cgi = argv[ii];
			continue;
		}
		if (strcmp(argv[ii], "--stksize") == 0) {
			if (++ii < argc)
				stksize = atoi(argv[ii]);
			continue;
		}

		dph_usage(argv[0]);
		return 1;
	}

	signal(SIGCHLD, dph_sigchld);
	signal(SIGPIPE, SIG_IGN);

	if (dph_init() == -1) {

		return 2;
	}

	if (cgi) {
		args[0] = cgi;
		args[1] = NULL;

		for (ii = 0; ii < ncgis; ii++)
			dph_createcgi(args, dph_pipe_session);
	} else {
		for (ii = 0; ii < ncgis; ii++)
			dph_createpipetest(256, 250, 8, dph_pipe_session);
	}

	while (numfds > 0)
		dph_scheduler(0, STD_SCHED_TIMEOUT);

	dph_cleanup();
	return 0;
}



#endif	/* #if defined(PIPETESTER) */

